﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace howto_path_through_array
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // The numbers.
        private int[,] Numbers = null;
        private int NumRows = 0;
        private int NumCols = 0;

        // The best path.
        private int[] BestPath = null;
        private int BestTotal = int.MinValue;

        private void btnMakeNumbers_Click(object sender, EventArgs e)
        {
            NumRows = (int)nudNumRows.Value;
            NumCols = (int)nudNumCols.Value;
            Numbers = new int[NumRows, NumCols];
            Random rand = new Random();
            for (int row = 0; row < NumRows; row++)
            {
                for (int col = 0; col < NumCols; col++)
                {
                    Numbers[row, col] = rand.Next(1, 99);
                }
            }
            BestPath = null;

            // Display the numbers.
            DisplayNumbers();

            // Enable the Find Path button.
            btnFindPath.Enabled = true;
            AcceptButton = btnFindPath;
            lblTotal.Text = "";
        }

        // Display the numbers and the best path.
        private void DisplayNumbers()
        {
            // Display the numbers.
            string txt = "";
            for (int row = 0; row < NumRows; row++)
            {
                txt += Environment.NewLine;
                for (int col = 0; col < NumCols; col++)
                {
                    txt += string.Format("{0,3}", Numbers[row, col]);
                }
            }
            txt = txt.Substring(Environment.NewLine.Length);
            rchNumbers.Text = txt;

            // Display the best path.
            if (BestPath == null) return;
            rchNumbers.Select();
            rchNumbers.SelectionBackColor = Color.White;
            for (int row = 0; row < NumRows; row++)
            {
                int start = 1 +
                    row * (NumCols * 3 + Environment.NewLine.Length - 1) +
                    BestPath[row] * 3;
                rchNumbers.Select(start, 2);
                rchNumbers.SelectionBackColor = Color.LightGreen;
            }
            rchNumbers.Select(0, 0);

            // Display the best path in the Console window.
            for (int row = 0; row < NumRows; row++)
                Console.Write(Numbers[row, BestPath[row]] + " ");
            Console.WriteLine("");
        }

        // Find the best path through the numbers.
        private void btnFindPath_Click(object sender, EventArgs e)
        {
            lblTotal.Text = "";
            Cursor = Cursors.WaitCursor;

            // Make one BestPath entry per row.
            BestPath = new int[NumRows];
            BestTotal = int.MinValue;

            // Find the best path.
            int total = FindBestPath();
            lblTotal.Text = total.ToString();

            // Redisplay the numbers to show the solution.
            DisplayNumbers();

            Cursor = Cursors.Default;
        }

        // Find the best path.
        // Return the path's total value.
        private int FindBestPath()
        {
            // Make a test path, one entry per row.
            int[] test_path = new int[NumRows];

            // Start in row 0 and try each column.
            for (int col = 0; col < NumCols; col++)
            {
                test_path[0] = col;
                FindPath(test_path, 1);
            }

            // Return the best total.
            return BestTotal;
        }

        // Recursively explore paths starting with the indicated row.
        // Return the path's total value.
        private void FindPath(int[] test_path, int row)
        {
            // See if we have a complete solution.
            if (row >= NumRows)
            {
                // We have a complete solution.
                // See if it is better than the best solution so far.
                int total = 0;
                for (int r = 0; r < NumRows; r++) total += Numbers[r, test_path[r]];
                if (total > BestTotal)
                {
                    // This is an improvement.
                    BestTotal = total;
                    Array.Copy(test_path, BestPath, NumRows);
                }
            }
            else
            {
                // We don't have a complete solution.
                // Try the possibilities for this row.

                // Get the column used in the previous row.
                int prev_col = test_path[row - 1];

                // Column - 1
                if (prev_col > 0)
                {
                    test_path[row] = prev_col - 1;
                    FindPath(test_path, row + 1);
                }

                // Column
                test_path[row] = prev_col;
                FindPath(test_path, row + 1);

                // Column + 1
                if (prev_col < NumCols - 1)
                {
                    test_path[row] = prev_col + 1;
                    FindPath(test_path, row + 1);
                }
            }
        }
    }
}
